from importlib import import_module
from threading import local
from warnings import warn

import click
from dramatiq import (
    Middleware,
    actor as register_actor,
    set_broker,
)
from dramatiq.cli import (
    CPUS,
    main as dramatiq_worker,
    make_argument_parser as dramatiq_argument_parser,
)
from flask import current_app
from flask.cli import with_appcontext


def import_object(path):
    # Implement setuptools entrypoint-like loading of object.
    modname, objname = path.split(':')
    mod = import_module(modname)
    try:
        return getattr(mod, objname)
    except AttributeError:
        raise ImportError("%s does not exists." % path)


class AppContextMiddleware(Middleware):
    # Setup Flask app for actor. Borrowed from
    # https://github.com/Bogdanp/flask_dramatiq_example.

    state = local()

    def __init__(self, app):
        self.app = app

    def before_process_message(self, broker, message):
        context = self.app.app_context()
        context.push()

        self.state.context = context

    def after_process_message(
            self, broker, message, *, result=None, exception=None):
        try:
            context = self.state.context
            context.pop(exception)
            del self.state.context
        except AttributeError:
            pass

    after_skip_message = after_process_message


class Dramatiq:
    # The Flask extension.

    # Reuse same defaults as dramatiq. cf.
    # https://github.com/Bogdanp/dramatiq/blob/master/dramatiq/broker.py#L34-L44
    DEFAULT_BROKER = 'dramatiq.brokers.rabbitmq:RabbitmqBroker'

    def __init__(self, app=None, broker_cls=DEFAULT_BROKER, name='dramatiq',
                 config_prefix=None):
        self.actors = []
        self.app = None
        self.broker_cls = broker_cls
        self.config_prefix = config_prefix or name.upper() + '_BROKER'
        self.name = name
        if app:
            self.init_app(app)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def init_app(self, app):
        if self.app is not None:
            warn(
                "%s is used by more than one flask application. "
                "Actor's context may be set incorrectly." % (self,),
                stacklevel=2,
            )
        self.app = app
        app.extensions['dramatiq-' + self.name] = self
        app.config.setdefault(self.config_prefix, self.broker_cls)
        cls = app.config[self.config_prefix]
        if isinstance(cls, str):
            cls = import_object(cls)
        kw = {}
        url = app.config.get(self.config_prefix + '_URL')
        if url:
            kw['url'] = url
        self.broker = cls(**kw)
        self.broker.add_middleware(AppContextMiddleware(app))

        for actor in self.actors:
            actor.register(broker=self.broker)

    def actor(self, fn=None, **kw):
        # Substitude dramatiq.actor decorator to return a lazy wrapper. This
        # allow to register actors in extension before the broker is
        # effectively configured by init_app.

        def decorator(fn):
            lazy_actor = LazyActor(self, fn, kw)
            self.actors.append(lazy_actor)
            if self.app:
                lazy_actor.register(self.broker)
            return lazy_actor

        if fn:
            return decorator(fn)
        return decorator


class LazyActor(object):
    # Intermediate object that register actor on broker an call.

    def __init__(self, extension, fn, kw):
        self.extension = extension
        self.fn = fn
        self.kw = kw
        self.actor = None

    def __call__(self, *a, **kw):
        return self.fn(*a, **kw)

    def __getattr__(self, name):
        if not self.actor:
            raise AttributeError(name)
        return getattr(self.actor, name)

    def register(self, broker):
        self.actor = register_actor(broker=broker, **self.kw)(self.fn)

    # Next is regular actor API.

    def send(self, *a, **kw):
        return self.actor.send(*a, **kw)

    def send_with_options(self, *a, **kw):
        return self.actor.send_with_options(*a, **kw)


@click.command()
@click.option('-p', '--processes', default=CPUS,
              metavar='PROCESSES', show_default=True,
              help="the number of worker processes to run")
@click.option('-t', '--threads', default=8,
              metavar='THREADS', show_default=True,
              help="the number of worker treads per processes")
@click.option('-Q', '--queues', type=str, default=None,
              metavar='QUEUES', show_default=True,
              help="listen to a subset of queues, comma separated")
@click.argument('broker_name', default='dramatiq')
@with_appcontext
def worker(processes, threads, queues, broker_name):
    """Run dramatiq workers.

    Setup Dramatiq with broker and task modules from Flask app.

    \b
    examples:
      # Run dramatiq with 1 thread per process.
      $ flask worker --threads 1

    \b
      # Listen only to the "foo" and "bar" queues.
      $ flask worker -Q foo,bar

    \b
      # Consuming from a specific broker
      $ flask worker mybroker
    """
    # Plugin for flask.commands entrypoint.
    #
    # Wraps dramatiq worker CLI in a Flask command. This is private API of
    # dramatiq.

    parser = dramatiq_argument_parser()

    # Set worker broker globally.
    needle = 'dramatiq-' + broker_name
    set_broker(current_app.extensions[needle].broker)

    command = [
        "--processes", str(processes),
        "--threads", str(threads),
        # This module does not have broker local. Thus dramatiq fallbacks to
        # global broker.
        __name__,
    ]
    if queues:
        command += ["--queues"] + queues.split(",")
    args = parser.parse_args(command)
    dramatiq_worker(args)



# -*- coding: utf-8 -*-
from distutils.core import setup

modules = \
['flask_dramatiq']
install_requires = \
['dramatiq>=1.5,<2.0']

entry_points = \
{'flask.commands': ['worker = flask_dramatiq:worker']}

setup_kwargs = {
    'name': 'flask-dramatiq',
    'version': '0.3.2',
    'description': 'Adds Dramatiq support to your Flask application',
    'long_description': '![Flask-Dramatiq](https://gitlab.com/bersace/flask-dramatiq/raw/master/docs/logo-horizontal.png?inline=false)\n\nFlask-Dramatiq plugs [Dramatiq](https://dramatiq.io) task queue in your\n[Flask](https://flask.pocoo.org) web application.\n\n## // Features // \n\n- Configure Dramatiq from Flask configuration.\n- Ensure Flask app is available to Dramatiq actor.\n- Add `worker` command to Flask CLI.\n- Enable [Flask Application factory](http://flask.pocoo.org/docs/dev/tutorial/factory/).\n- Handle multiple brokers with configurable prefix.\n\nFull documentation at\n[flask-dramatiq.readthedocs.io](https://flask-dramatiq.readthedocs.io).\n\n\n## // Installation and Usage //\n\nFlask-Dramatiq is licensed under BSD-3-Clause. Add `flask-dramatiq` to your\nproject:\n\n``` console\n$ poetry add flask-dramatiq\n```\n\nThen use `Dramatiq` object as a regular Flask extension:\n\n``` python\nfrom flask import Flask\nfrom flask_dramatiq import Dramatiq\n\napp = Flask(__name__)\ndramatiq = Dramatiq(app)\n\n@dramatiq.actor()\ndef my_actor():\n    ...\n\n@app.route("/")\ndef myhandler():\n    my_actor.send()\n```\n\nFlask-Dramatiq adds two configuration keys:\n\n- `DRAMATIQ_BROKER`, points to broker class like\n  `dramatiq.brokers.rabbitmq.RabbitmqBroker` or\n  `dramatiq.brokers.redis.RedisBroker`.\n- `DRAMATIQ_BROKER_URL` is passed as `url` keyword argument to broker class.\n\nNow run worker program to consume messages and execute tasks in the background:\n\n``` console\n$ flask worker --processes=1\n```\n\nA complete flask app is available in project source tree\n[example.py](https://gitlab.com/bersace/flask-dramatiq/blob/master/example.py).\n\n\n## // Credit and Support //\n\nFeel free to open an issue or suggest a merge request on [Gitlab project\npage](https://gitlab.com/bersace/flask-dramatiq). Contribution welcome!\n\nThe project is based on\n[Bogdanp/flask_dramatiq_example](https://github.com/Bogdanp/flask_dramatiq_example).\n',
    'author': 'Étienne BERSAC',
    'author_email': None,
    'url': 'https://pypi.org/project/flask-dramatiq',
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
